import logging

from .class_action_failed import ActionFailed
from .class_state import State


logger = logging.getLogger(__name__)


class Action(object):
    """Represents an executable action.
    """

    def __init__(self, func, *args, **kwargs):
        """Instantiates a new executable action.

        Args:
            func (function): The funcion to execute. This function must be
            decorated with action.

            *args: Any additional arguments required by func.

            enabled (bool|State): A keyword argument that, if set to False,
            disables the Action from executing. A State instance can also be
            passed in for lazy evaluation. Defaults to True.

            pass_success (bool): A keyword argument that, if set to True,
            passes in the success of all other actions executed thus far as a
            boolean keyword argument to the func function.

            _finally (bool): A keyword argument that, if set to True, makes the
            action execute after all other actions have been executed. This
            will ensure the action is executed even if other actions failed.
            This is typically used for cleanup tasks. Return values from these
            actions are ignored. Defaults to False.

        Raises:
            ValueError: When the given function doesn't have a label set by the
            action decorator.
        """
        self._func = func
        self._args = args
        self._kwargs = {}
        self._enabled = kwargs.get("enabled", True)
        self._pass_success = bool(kwargs.get("pass_success", False))
        self._finally = bool(kwargs.get("_finally", False))

        if not hasattr(func, "label"):
            raise ValueError("%s is missing label. Make sure it uses the "
                "action decorator." % func.__name__)

        self.label = func.label

    def __call__(self, success, state, pre_action, post_action):
        """Perform the action.

        Args:
            success (bool): Whether or not the last action was successful or
            otherwise failed.
            state (dict): A group of values that have been saved by other
            Action instances.

        Raises:
            ActionFailed: When the Action fails during its execution.
        """
        if not self._is_enabled(state):
            logger.debug("Skipping '%s'", self.label)
            return True

        pre_action(self)

        if self._pass_success:
            self._kwargs["success"] = success

        args = self._inject_state_into_args(state, self._args)

        our_success = False
        try:
            our_success = self._func(*args, **self._kwargs)

        except Exception:
            logger.exception("Error running '%s'", self.label)

        if not our_success:
            raise ActionFailed(self.label)

        post_action(self)

    def _is_enabled(self, state):
        """Returns whether or not the Action is enabled.

        This lazy evaluates the enabled flag using the given state.

        Args:
            state (dict): A group of values that have been saved by other
            Action instances.

        Returns:
            bool: True if this Action is enabled. False otherwise.
        """
        enabled = True

        if isinstance(self._enabled, State):
            enabled = bool(state.get(
                self._enabled.name, self._enabled.default))

        else:
            enabled = bool(self._enabled)

        return enabled

    def _inject_state_into_args(self, state, args):
        """Lazy evaluates arguments by injecting state.

        Args:
            state (dict): The storage of state.
            args (list[mixed]): The list of arguments to inject state into.

        Returns:
            list[mixed]: The given args that have been lazy evaluated with
            injected state.
        """
        final_args = []
        for arg in args:
            if isinstance(arg, State):
                try:
                    final_args.append(state[arg.name])

                except KeyError:
                    raise KeyError("'%s' does not exist in saved state for "
                        "Action '%s'. Current state: %s" % (
                            arg.name, self.label, state))

            else:
                final_args.append(arg)

        return final_args



class ActionFailed(Exception):
    """Exception that's thrown when an Action fails.

    The label of the Action is stored publically in the `label` attribute.
    """

    def __init__(self, label):
        """Instantiates the instance.

        Args:
            label (str): The label of the Action that failed.
        """
        self.label = label



from functools import wraps


def action(label):
    """Decorator to convert a function to a callable usable by the Action
    class.

    Args:
        label (str): The label to show before executing the action.
    """

    def wrapper(f):
        """
        Returns:
            bool: Indicates whether or not the action was successful. Returns
            True if the function returned True (or does not have a return
            value). Otherwise returns False.
        """
        f.label = label

        @wraps(f)
        def wrap(*args, **kwargs):
            result = f(*args, **kwargs)

            if result is None:
                result = True

            return bool(result)

        return wrap

    return wrapper



class State(object):
    """Represents a value that was created in one StatefulAction and passed
    into another Action.

    This allows for Actions to be declarative and lazy evaluated.
    """

    def __init__(self, name, default=None):
        """Instantiates the instance.

        Args:
            name (str): The name of the value that this State represents.
            default (bool): The default value to use when the evaluated value
            is missing. Defaults to None.
        """
        self.name = name
        self.default = default



import logging

from .class_stateful_action import StatefulAction


logger = logging.getLogger(__name__)


def _null(action):
    pass


class Actions(object):
    """Represents a group of actions that performs tasks.
    """

    def __init__(self, *actions):
        """Instantiates a new group of actions.

        Args:
            actions (*Action): A list of actions to perform.
        """
        self._actions = [action for action in actions if not action._finally]
        self._finally = [action for action in actions if action._finally]
        self._state = {}
        self.success = False

    def __call__(self, pre_action=_null, post_action=_null):
        """Perform all actions.

        Args:
            pre_action (callable): An optional callable to call before an
            Action is executed. This function receives the Action as the first
            argument. Defaults to doing nothing.
            post_action (callable): An optional callable to call after an
            Action is executed. This function receives the Action as the first
            argument. Defaults to doing nothing.

        Raises:
            ActionFailed: When any Action fails during its execution.
        """
        self.success = False

        try:
            for action in self._actions:
                action(True, self._state, pre_action, post_action)
                self._save_state(action)

            self.success = True

        finally:
            for action in self._finally:
                try:
                    action(self.success, self._state, pre_action, post_action)
                    self._save_state(action)

                except Exception:
                    logger.exception("Failed running %s", action.label)

    def _save_state(self, action):
        if isinstance(action, StatefulAction):
            self._state.update(action.results)



import logging

from .action_decorator import action
from .class_action import Action
from .class_action_failed import ActionFailed
from .class_actions import Actions
from .class_state import State
from .class_stateful_action import StatefulAction


logging.getLogger(__name__).addHandler(logging.NullHandler())



import logging

import bunch

from .class_action import Action


logger = logging.getLogger(__name__)


class StatefulAction(Action):
    """Represents an executable action that stores state.

    When called, the `func` is passed in an extra keyword bunch.Bunch argument
    called `result_store`. (Bunch is a dict-like object where values can be set
    via attributes.) This can be used by the callable to store state.

    State can be passed onto other actions by referencing a State instance.
    """

    def __init__(self, func, *args, **kwargs):
        """Instantiates a new executable action. Same as Action parent class.

        Args:
            state (list[str]): Names of states that will be saved in the result
            store. After calling the StatefulAction action, these values will
            be retrievable via the results property.
        """
        self._expected_states = kwargs.get("state", [])
        self._result_store = bunch.Bunch()
        super(StatefulAction, self).__init__(func, *args, **kwargs)
        self._kwargs["result_store"] = self._result_store

    @property
    def results(self):
        """Retrieves all the state saved after calling the StatefulAction.

        If any expected state is missing, it defaults to None.

        Returns:
            dict: A dict representing the saved state of the StatefulAction.
        """
        state = {}

        for expected_state in self._expected_states:
            value = None

            if expected_state in self._result_store:
                value = self._result_store[expected_state]

            else:
                logger.debug("'%s' not found in result store for '%s' "
                    "StatefulAction", expected_state, self.label)

            state[expected_state] = value

        return state



"""
setup.py

Setup for the actions package
"""
import os

from setuptools import find_packages
from setuptools import setup


try:
    import pypandoc
    long_description = pypandoc.convert("README.md", "rst")

except (IOError, ImportError):
    long_description = ""


setup(
    name="actions",
    version="1.0.3",
    description="Declarative actions",
    long_description=long_description,
    author="Quid Inc.",
    author_email="ops@quid.com",
    url="https://github.com/quid/actions",
    packages=find_packages(),
    install_requires=[
        "bunch==1.0.1"
    ]
)
